package lucastest;

import java.util.*;

public class GTree {

    private int nTips;
    private GNode[] nodes;
    private SpeciesSet speciesSet;
    private Random rand;

    /**
     * This only creates the (fixed) tip nodes
     * @param config supplies nTipsA, nTipsB, nTipsC for tip assignments
     * @param speciesSet stored, for access to migration rate and coalescence rates, randomSTipForNode()
     * @param direct for deciding which seed for PRNG (direct vs mcmc)
     */
    public GTree(Config config, SpeciesSet speciesSet, boolean direct) {
        makeTipNodes(config);
        this.speciesSet = speciesSet;
        rand = new Random();
        rand.setSeed(direct ? config.directSampleSeed : config.initGTreeSeed);
    }


    public void makeTwoTipTree(double hgt) {
        assert nTips == 2;
        addNode(2, hgt, 0, 0, 1);
    }


    /**
     * Makes a random GTree
     *
     */
    public void makeSampleTree() {
        AllLineagesState als = new AllLineagesState(speciesSet.getNTips());
        als.currentLineages = new ArrayList<>();
        int n;
        for (n = 0; n < getNTips(); n++) {
            LineageState ls = new LineageState(n, getNode(n).getSTip());
            als.currentLineages.add(ls);
        }
        CoalOrMig coalOrMig = nextCoalOrMig(0.0, als, rand);
        while (n < getNNodes()) {
            n = applyNextCoalOrMig(coalOrMig, n, als, rand);
            coalOrMig = nextCoalOrMig(coalOrMig.hgt, als, rand);
        }
        fillinIDs();
    }



    /**
     *
     * @return a copy of the tree, for use by MCMC
     */
    public GTree getCopy() {
        GTree copy = new GTree();
        copy.nTips = getNTips();
        copy.nodes = new GNode[nodes.length];
        for (int i = 0; i < nodes.length; i++) {
            copy.nodes[i] = nodes[i].getCopy();
        }
        copy.speciesSet = speciesSet;
        copy.rand = rand;
        return copy;
    }


    public GNode getNode(int n) {
        assert n < nodes.length;
        return nodes[n];
    }


    public GNode getRoot() {
        assert nodes[nodes.length-1].getAnc() < 0;
        return nodes[nodes.length-1];
    }




    public String asText() {
        StringBuilder text = new StringBuilder("Gene tree\n");
        for (int n = 0; n < nodes.length; n++) {
            text.append(n).append("   ").append(nodes[n].asText()).append("\n");
        }
        return text.toString();
    }


    public String logColumnNames() {
        StringBuilder text = new StringBuilder("");
        for (int n = 0; n < nodes.length; n++) {
            text.append("ChX").append(n).append(" ");
            text.append("ChY").append(n).append(" ");
            text.append("anc").append(n).append(" ");
            text.append("hgt").append(n).append(" ");
            text.append("Stip").append(n).append(" ");
        }
        return text.toString();
    }


    public String logColumns() {
        StringBuilder text = new StringBuilder("");
        for (int n = 0; n < nodes.length; n++) {
            text.append(nodes[n].getChX()).append(" ");
            text.append(nodes[n].getChY()).append(" ");
            text.append(nodes[n].getAnc()).append(" ");
            text.append(String.format("%8.6f ", nodes[n].getHeight()));
            text.append(nodes[n].getSTip()).append(" ");
        }
        return text.toString();
    }


    class CoalIndexHeight {
        int n;
        double height;
        CoalIndexHeight(int n, double height) {
            this.n = n;
            this.height = height;
        }
        public double getHeight() {
            return height;
        }
    }

    public int [] coalsInHeightOrder() {
        ArrayList<CoalIndexHeight> coalIHs = new ArrayList<>();
        for (int n = getNTips(); n < getNNodes(); n++) {
            CoalIndexHeight cHI = new CoalIndexHeight(n, getNode(n).getHeight());
            coalIHs.add(cHI);
        }
        Collections.sort(coalIHs,  Comparator.comparingDouble(CoalIndexHeight::getHeight));
        int [] indices = new int[coalIHs.size()];
        for (int i = 0; i < coalIHs.size(); i++) {
            indices[i] = coalIHs.get(i).n;
        }
        return indices;
    }


    ArrayList<Integer> collectPersistersAtCoal(int i, int j, double coalT) {
        ArrayList<Integer> otherLins = new ArrayList<>();
        for (int n = 0; n < getNNodes(); n++) {
            int anc = getNode(n).getAnc();
            if (anc >= 0  && n != i  && n != j && getNode(n).getHeight() < coalT  &&  getNode(anc).getHeight() > coalT) {
                otherLins.add(n);
            }
        }
        return otherLins;
    }

    public int getNTips() {
        return nTips;
    }


    public int getNNodes() { return nodes.length; }



    /******************************************************************************************************************/
    /**********************************           private        ******************************************************/
    /******************************************************************************************************************/

    private static class LineageState {
        int gIdx;
        int sIdx;

        LineageState(int gIdx, int sIdx) {
            this.gIdx = gIdx;
            this.sIdx = sIdx;
        }
    }


    private static class AllLineagesState {
        int nSpBranches;
        ArrayList<LineageState> currentLineages;

        AllLineagesState(int nSpBranches) {
            this.nSpBranches = nSpBranches;
        }
    }


    private static class CoalOrMig {
        // If a coal, i and j are gtree node indices, both >= 0,
        int i;
        int j;
        // if a mig, i<0, j<0, and lin is currentLineages index, b destination branch
        int lin;
        int b;
        double hgt;
    }


    private void makeTipNodes(Config config) {
        nTips = 0;
        for (int nGtipsAtStip : config.nGtreeTips) {
            nTips += nGtipsAtStip;
        }
        assert nTips < 26;
        nodes = new GNode[2*nTips-1];
        int n = 0;
        for (int sTip = 0; sTip < config.nGtreeTips.size(); sTip++) {
            for (int i = 0; i < config.nGtreeTips.get(sTip); i++) {
                nodes[n] = new GNode("abcdefghijklmn0pqrstuvwxyz".substring(n, n+1), 0.0, sTip);
                n++;
            }
        }
        assert n==nTips;
    }



    // for getCopy()
    private GTree() {
    }



    private int addNode(int n, double hgt, int sTip, int chX, int chY) {
        nodes[n] = new GNode("", hgt, sTip);
        nodes[n].setChX(chX);
        nodes[n].setChY(chY);
        nodes[chX].setAnc(n);
        nodes[chY].setAnc(n);
        return n+1;
    }



    private CoalOrMig nextCoalOrMig(double hgt, AllLineagesState als, Random rand) {
        double mRate = speciesSet.getMRate();
        int nLins = als.currentLineages.size();
        double[][] mRates = new double[nLins][als.nSpBranches];
        double[][] cRates = new double[nLins][nLins];
        double totalRate = 0.0;
        for (int i = 0; i < nLins; i++) {
            int b = als.currentLineages.get(i).sIdx;
            double coalRate = speciesSet.getCRate(b);
            for (int j = i + 1; j < nLins; j++) {
                if (b == als.currentLineages.get(j).sIdx) {
                    cRates[i][j] = coalRate;
                    totalRate += coalRate;
                }
            }
            for (int c = 0; c < als.nSpBranches; c++) {
                if (c != b) {
                    mRates[i][c] = mRate;
                    totalRate += mRate;
                }
            }
        }
        double newHgt = hgt - Math.log(rand.nextDouble()) / totalRate;
        CoalOrMig coalOrMig = new CoalOrMig();
        coalOrMig.hgt = newHgt;
        double partialSum = 0.0;
        double xx = totalRate * rand.nextDouble();
        for (int lin1 = 0; lin1 < nLins; lin1++) {
            for (int lin2 = lin1 + 1; lin2 < nLins; lin2++) {
                partialSum += cRates[lin1][lin2];
                if (partialSum > xx) {
                    coalOrMig.i = als.currentLineages.get(lin1).gIdx;
                    coalOrMig.j = als.currentLineages.get(lin2).gIdx;
                    coalOrMig.lin = -1;
                    coalOrMig.b = als.currentLineages.get(lin1).sIdx;
                    return coalOrMig;
                }
            }
            for (int c = 0; c < als.nSpBranches; c++) {
                partialSum += mRates[lin1][c];
                if (partialSum > xx) {
                    coalOrMig.i = -1;
                    coalOrMig.j = -1;
                    coalOrMig.lin = lin1;
                    coalOrMig.b = c;
                    return coalOrMig;

                }
            }
        }
        assert false;
        return null;
    }


    private int  applyNextCoalOrMig(CoalOrMig coalOrMig, int n, AllLineagesState als, Random rand) {
        if (coalOrMig != null) {
            if (coalOrMig.i >= 0) {
                assert coalOrMig.j >= 0;
                addNode(n, coalOrMig.hgt, coalOrMig.b, coalOrMig.i, coalOrMig.j);
                als.currentLineages.removeIf(lineageState -> (lineageState.gIdx == coalOrMig.i || lineageState.gIdx == coalOrMig.j));
                als.currentLineages.add(new LineageState(n, coalOrMig.b));
                n++;
            } else {
                assert coalOrMig.lin >= 0;
                als.currentLineages.get(coalOrMig.lin).sIdx = coalOrMig.b;
            }
        }
        return n;
    }




    private void fillinIDsSubtree(int n) {
        int chX = nodes[n].getChX();
        int chY = nodes[n].getChY();
        if (chX >= 0) {
            fillinIDsSubtree(chX);
            fillinIDsSubtree(chY);
            nodes[n].setId(nodes[chX].getID() + nodes[chY].getID());
        }
    }


    private void fillinIDs() {
        assert nodes[nodes.length-1].getAnc() < 0;
        fillinIDsSubtree(nodes.length-1);
    }


}
